﻿using System;
using System.Collections.Generic;
using System.Text;

namespace ConsoleApp2
{
    public class Extension
    {
        public static IEnumerable<int> Range(int start, int stop, int step = 1)
        {
            if (step == 0)
                throw new ArgumentException("Parameter step cannot equal zero.");

            if (start < stop && step > 0)
            {
                for (var i = start; i < stop; i += step)
                {
                    yield return i;
                }
            }
            else if (start > stop && step < 0)
            {
                for (var i = start; i > stop; i += step)
                {
                    yield return i;
                }
            }
        }

        /*

        // Is there a constraint that restricts my generic method to numeric types?
        // https://stackoverflow.com/questions/32664/is-there-a-constraint-that-restricts-my-generic-method-to-numeric-types
        public static IEnumerable<T> RangePython<T>(T start, T stop, T step)
        {
            dynamic dStart = start;
            dynamic dStop = stop;
            dynamic dStep = step;

            if (dStep == 0)
                throw new ArgumentException("Parameter step cannot equal zero.");

            if (dStart < dStop && dStep > 0)
            {
                for (var i = dStart; i < dStop; i += dStep)
                {
                    yield return i;
                }
            }
            else if (dStart > dStop && dStep < 0)
            {
                for (var i = dStart; i > dStop; i += dStep)
                {
                    yield return i;
                }
            }
        }
        */


        /*
        public static IEnumerable<T> Range<T>(T start, T stop, T step = 1) where T : unmanaged
        {
            if (step == 0)
                throw new ArgumentException("Parameter step cannot equal zero.");

            if (start < stop && step > 0)
            {
                for (var i = start; i < stop; i += step)
                {
                    yield return i;
                }
            }
            else if (start > stop && step < 0)
            {
                for (var i = start; i > stop; i += step)
                {
                    yield return i;
                }
            }
        }
        */

        /*
        public static IEnumerable<T> RangePython<T>(T start, T stop, T step) where T : unmanaged, IComparable, IEquatable<T>
        {
            if (step.Equals(default(T)))
                throw new ArgumentException("Parameter step cannot equal zero.");

            if (start.CompareTo(stop) < 0 && step.CompareTo(0) > 0)
            {
                for (var i = start; i.CompareTo(stop) < 0; i += step)
                {
                    yield return i;
                }
            }
            else if (start.CompareTo(stop) > 0 && step.CompareTo(0) < 0)
            {
                for (var i = start; i.CompareTo(stop) > 0; i += step)
                {
                    yield return i;
                }
            }
        }
        */

        public static IEnumerable<T> RangePython<T>(T start, T stop, T step) where T : unmanaged, IComparable, IEquatable<T>
        {
            if (step.Equals(default(T)))
                throw new ArgumentException("Parameter step cannot equal zero.");

            if (start.CompareTo(stop) < 0 && step.CompareTo(default(T)) > 0)
            {
                for (var i = start; i.CompareTo(stop) < 0;)
                {
                    yield return i;
                    Add(ref i, step);
                }
            }
            else if (start.CompareTo(stop) > 0 && step.CompareTo(default(T)) < 0)
            {
                for (var i = start; i.CompareTo(stop) > 0;)
                {
                    yield return i;
                    Add(ref i, step);
                }
            }
        }

        private unsafe static void Add<T>(ref T value, T step) where T : unmanaged
        {
            switch (value)
            {
                case short s:
                    short sStep = (short)(object)step;
                    s += sStep;
                    value = (T)(object)s;
                    break;

                case int i:
                    int iStep = (int)(object)step;
                    i += iStep;
                    value = (T)(object)i;
                    break;

                case double _:
                    {
                        double* pStep = (double *)&step;
                        fixed (T* pValue = &value)
                        {
                            double* pTargetDouble = (double*)pValue;
                            *pTargetDouble += *pStep;
                        }
                    }
                    break;

                default:
                    throw new NotSupportedException(value.GetType().ToString());
            }
        }

        /*
        private unsafe static void Add<T>(ref T i, T step) where T : unmanaged, IComparable, IEquatable<T>
        {
            T* pStep = &step;
            fixed (T* pValue = &i)
            {
                switch (i)
                {
                    case short _:
                        short* pTargetShort = (short*)pValue;
                        short* pIncShort = (short*)pStep;
                        *pTargetShort += *pIncShort;
                        break;

                    case int _:
                        int* pTargetInt = (int*)pValue;
                        int* pIncInt = (int*)pStep;
                        *pTargetInt += *pIncInt;
                        break;

                    case double _:
                        double* pTargetDouble = (double*)pValue;
                        double* pIncDouble = (double*)pStep;
                        *pTargetDouble += *pIncDouble;
                        break;

                    default:
                        throw new NotSupportedException(i.GetType().ToString());
                }
            }
        }
        */
    }
}
